package org.implab.gradle.containers.tasks;

import java.io.FileInputStream;
import java.io.IOException;
import java.io.InputStream;
import java.nio.file.Files;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Optional;

import org.apache.tools.ant.taskdefs.UpToDate;
import org.gradle.api.Action;
import org.gradle.api.file.DirectoryProperty;
import org.gradle.api.file.RegularFile;
import org.gradle.api.file.RegularFileProperty;
import org.gradle.api.provider.ListProperty;
import org.gradle.api.provider.MapProperty;
import org.gradle.api.provider.Property;
import org.gradle.api.tasks.Input;
import org.gradle.api.tasks.InputDirectory;
import org.gradle.api.tasks.OutputFile;
import org.gradle.api.tasks.SkipWhenEmpty;
import org.gradle.api.tasks.TaskAction;
import org.implab.gradle.containers.ImageName;
import org.implab.gradle.containers.cli.Utils;
import org.implab.gradle.containers.dsl.MapPropertyEntry;

import groovy.lang.Closure;

public abstract class BuildImage extends DockerCliTask {

    public final String BUILD_COMMAND = "build";

    @InputDirectory
    @SkipWhenEmpty
    public abstract DirectoryProperty getContextDirectory();

    @Input
    public abstract MapProperty<String, String> getBuildArgs();

    @Input
    public abstract ListProperty<String> getExtraCommandArgs();

    @Input
    @org.gradle.api.tasks.Optional
    public abstract Property<String> getBuildTarget();

    @Input
    public abstract Property<ImageName> getImageName();

    @OutputFile
    public abstract RegularFileProperty getImageIdFile();

    protected BuildImage() {
        getOutputs().upToDateWhen(task -> getImageIdFile()
                .map(RegularFile::getAsFile)
                .map(docker()::imageExists)
                .getOrElse(false));
    }

    public void buildArgs(Action<Map<String, String>> spec) {
        getBuildArgs().putAll(provider(() -> {
            Map<String, String> args = new HashMap<>();
            spec.execute(args);
            return args;
        }));
    }

    public void buildArgs(Closure<Map<String, String>> spec) {
        buildArgs(Utils.wrapClosure(spec));
    }

    public MapPropertyEntry<String, String> buildArg(String key) {
        return new MapPropertyEntry<String, String>(getBuildArgs(), key, getProviders());
    }

    @TaskAction
    public void Run() throws Exception {
        List<String> args = new ArrayList<>();

        args.addAll(List.of(
                "-t", getImageName().get().toString(),
                "--iidfile", getImageIdFile().getAsFile().get().toString()));

        getBuildArgs().get().forEach((k, v) -> {
            args.add("--build-arg");
            args.add(String.format("%s=%s", k, v));
        });

        // add --target if specified for multi-stage build
        if (getBuildTarget().isPresent()) {
            args.add("--target");
            args.add(getBuildTarget().get());
        }

        // add extra parameters
        getExtraCommandArgs().getOrElse(Collections.emptyList())
                .forEach(args::add);

        args.add(getContextDirectory().getAsFile().get().toString());

        docker().build(args);
    }


    @Override
    protected Optional<String> getSubCommand() {
        return Optional.of(BUILD_COMMAND);
    }

    @Override
    protected Collection<String> getSubCommandArguments() {
        List<String> args = new ArrayList<>();

        args.addAll(List.of(
                "-t", getImageName().get().toString(),
                "--iidfile", getImageIdFile().getAsFile().get().toString()));

        getBuildArgs().get().forEach((k, v) -> {
            args.add("--build-arg");
            args.add(String.format("%s=%s", k, v));
        });

        // add --target if specified for multi-stage build
        if (getBuildTarget().isPresent()) {
            args.add("--target");
            args.add(getBuildTarget().get());
        }

        // add extra parameters
        getExtraCommandArgs().getOrElse(Collections.emptyList())
                .forEach(args::add);

        args.add(getContextDirectory().getAsFile().get().toString());

        return args;
    }
}
